#ifndef __ED_LIST_H__
#define __ED_LIST_H__

#include <iostream>
#include <editable/editableattribute.h>
#include <vector>

/** @class EdList
 * @brief Editable attribute for lists
 * Each entry of the list is again a list of attributes.
 * This kind of complex attribute was thought for things like editting the
 * switched state for each osg::Switch child, or the min/max range for each
 * osg::Lod child and so on.
 * Each entry is in fact very similar to an EditableEntity, but the entry
 * doesn't have entity, it's only a list of attributes binded to some
 * index, without even a name.
 */
class EdList: public EditableAttribute {
public:
    typedef std::pair< std::string, EditableAttributeList > ListItem;
    typedef std::vector< ListItem > List;
public:
    EdList(std::string name, std::string description);

    /** Returns the type of this attribute */
    virtual std::string getType() { return "list"; }

    /** Get access to the name of one of the attributes */
    inline const std::string getItemName(unsigned i) { return values[i].first; }

    /** Get access to one of the attributes (file) */
    inline const EditableAttributeList& get(unsigned i) { return values[i].second; }

    /** Get access to one concrete attribute (file,column) */
    inline EditableAttribute *get(unsigned row, unsigned column) {
	if (values.size() > row) {
	    if (values[row].second.size() > column) {
		return values[row].second[column].get();
	    } else {
		std::cerr << "FATAL ERROR[col]: " << column << " > " << values[row].second.size() << std::endl;
	    }
	} else {
	    std::cerr << "FATAL ERROR[row]: " << row << " > " << values.size() << std::endl;
	}
	return 0;
    }

    inline unsigned length() { return values.size(); }

    /** Take care of add all the attributes of the same type, or this method
     * will fail */
    void add(std::string name, EditableAttributeList entry) { values.push_back(ListItem(name, entry)); }

    /** If you want to allow addition/removal of items to the list, you must
     * supply a prototype of item (an example of the list of attributes that
     * can be in a row) */
    void allowAddRemove(EditableAttributeList prototype) { this->prototype_item=prototype; }

    /** If this class has a prototype that allows addition of new items, calling
     * this method a new item is created.
     * @return The index in the list of the new added attribute */
    unsigned addNew();

    /** If addition/removal is enabled, this method removes one item from the
     * list */
    void remove(unsigned i);

    /** Clears the entire list. Take care, use this *ONLY* if you will change the
     * contents of the list for another ones. Don't leave an empty list!
     */
    void clear() { while (values.size()) values.erase(values.begin()); }

    /** Checks if somebody has enabled the addition/removal of items */
    bool canAddRemove() const { return prototype_item.size() > 0; }

    /** Access to the prototype */
    inline const EditableAttributeList& getPrototype() { return prototype_item; }

    /** Checks if the attribute has changed */
    virtual bool hasChanged() const;

    inline virtual std::string toString() const {
	std::cerr << "FATAL ERROR (BUG)!: Asking string value of a List!!!" << std::endl;
	throw -1;
    }

    virtual EditableAttribute *clone() const ;
private:
    List values;
    EditableAttributeList prototype_item;
};

#endif
