#ifndef __ED_REFERENCE_H__
#define __ED_REFERENCE_H__

#include <editable/editableattribute.h>
#include <map>

class EditableEntity;

/** @class EdReference
 * @brief Editable attribute for references to entities
 *
 * It has a binded list of classes, to understand the meaning of this,
 * see an example:
 *  Imagine that you have a class A with a reference to a class B. But
 *  this class B is abstract, with two concrete subclasses C, D. Now
 *  you want to give the user the posibility of change either the attributes
 *  of the actual referenced object, or even change the referenced object
 *  itself, to be of another class. That is, imagine, that your instance of
 *  A (a), has a reference to an instance of C (c), but user wants to change
 *  a to use an instance of D instead. This is the reason why EdReference has
 *  a list of classes, they are in fact prototypes of all the possible
 *  classes allowed. This way, you can change this reference to use another
 *  referenced class. Polymorphism is good ;)
 *
 *  A more concrete example: You have class Employ, which has a reference
 *  to an Enterprise. But Enterprise is abstract, with subclasses Provider
 *  and Client. Now you open your application and you find that some Employ
 *  is binded to a Provider, when in fact should be binded with a Client.
 *  You can ask EdReference to point to an instance of Client.
 *
 * Anyway, it's possible to have a reference to some instance and don't allow
 * changing it. Simply don't register classes. User wan't be able to change
 * the class of the referenced object. The interface should reflect this by
 * disabling the buttons used for this task.
 */
class EdReference: public EditableAttribute {
public:
    EdReference(std::string name, std::string description);

    /** Returns the type of this attribute */
    virtual std::string getType() { return "reference"; }

    /** Get the current referenced entity */
    inline EditableEntity* get() { return value.get(); }

    /** Get the original entity before any user change */
    inline EditableEntity* getOriginal() { return original.get(); }

    /// Adds a new class supported by this reference
    void addClass(std::string name, EditableEntity *prototype);

    /// Instantiates one of the added prototypes as the actual reference
    void set(std::string the_class);

    /// Sets initial pointer
    void setInitial(EditableEntity *entity);

    /// Sets the reference to null
    void setNull();

    /** Checks if the attribute has changed */
    inline virtual bool hasChanged() const { return original != value; }

    /** Get an string representation of this attribute */
    virtual std::string toString() const;

    /** Obtains the list of classes, that is, the possible classes to which
     * this reference can point, to allow the interface to select one of them */
    std::vector<std::string> getClasses();

    /** Obtains the number of possible classes */
    inline unsigned getNumClasses() { return map.size(); }

    virtual EditableAttribute *clone()const ;

    /** Here you can specify if it will be allowed to change the reference
     * to point to another object or NULL. Take care, it has no sense to
     * forbid the change and register several classes, as the user won't be
     * able to change between them. If you forbid changing, specify at most
     * one class. */
    void setAllowChange(bool allow) { this->allow_change = allow; }

    /** Here you can test if it will be allowed to change the reference
     * to point to another object or NULL */
    bool getAllowChange() { return allow_change; }
private:
    osg::ref_ptr<EditableEntity> value; ///< Current value
    osg::ref_ptr<EditableEntity> original; ///< Original value
    std::map<std::string, osg::ref_ptr<EditableEntity> > map;
    bool allow_change; ///< Allow changing the reference to point to another objet
};

#endif
