#ifndef __EDITABLE_ENTITY_H__
#define __EDITABLE_ENTITY_H__

#include <osg/Object>
#include <string>
#include "editableattribute.h"
#include <command/command.h>

/** @class EditableEntity
 * @brief Editable class of OSG
 *
 * This is the interface that may follow all the wrappers that make OSG
 * classes editable.
 */
class EditableEntity: public osg::Referenced {
public:
    EditableEntity() { inheritedCommit = false; }
    /** Commits all the changes in the OSG class creating and executing a
     * command */
    virtual std::string getName() const =0;

    /// Commits the changes, generating a command and executing it
    virtual void commit()=0;

    /** Compiles the changes into a command, used for example in composite
     * entities, when one need to build a meta-command with all the changes
     * to be able to undo/redo all the changes in group */
    virtual Command *compile()=0;

    /// Sets the actual entity edited that is wrapped by this class
    virtual void setObject(osg::Object *object)=0;

    /// Returns the actual entity edited
    virtual osg::Object *getObject()=0;

    /// Generates a new entity to be edited that is wrapped by this class
    virtual void genObject()=0;

    /// Returns the list of attributes of this entity
    virtual EditableAttributeList getAttributes()=0;

    /// Prototype pattern
    virtual EditableEntity *clone() const =0;

    /** Sets whether or not the commit will be inherited (ie, called from
     * another entity that contains this one). */
    inline void setInheritedCommit(bool b) { inheritedCommit=b; }

    /** Gets whether or not the commit will be inherited (ie, called from
     * another entity that contains this one). It may be usefull for the
     * interface, to know if it should allow the user to commit this single
     * entity or not. */
    inline bool getInheritedCommit() { return inheritedCommit; }
private:
    /** Indicates if the commit will be inherited (ie, called from
     * another entity that contains this one). False by default. */
    bool inheritedCommit; 
};

#endif //__EDITABLE_ENTITY_H__
