#include <osg/Texture2D>
#include <osg/TextureCubeMap>
#include <editable/entityfactory.h>

#include <editable/entities/edshademodel.h>
#include <editable/entities/edmaterial.h>
#include <editable/entities/edtexture.h>
#include <editable/entities/edblendfunc.h>
#include <editable/entities/edcullface.h>
#include <editable/entities/edfrontface.h>
#include <editable/entities/edlightmodel.h>
#include <editable/entities/edfog.h>
#include <editable/entities/edtexmat.h>
#include <editable/entities/edvertexprogram.h>


#include "edstateset.h"

#include <command/comstateset.h>

static std::string getGLName(osg::StateAttribute::GLMode mode) {
    switch (mode) {
    case GL_ALPHA_TEST: return "GL_ALPHA_TEST";
    case GL_BLEND: return "GL_BLEND";
    case GL_COLOR_MATERIAL: return "GL_COLOR_MATERIAL";
    case GL_CULL_FACE: return "GL_CULL_FACE";
    case GL_DEPTH_TEST: return "GL_DEPTH_TEST";
    case GL_FOG: return "GL_FOG";
    case GL_LIGHTING: return "GL_LIGHTING";
    case GL_POINT_SMOOTH: return "GL_POINT_SMOOTH";
    case GL_LINE_STIPPLE: return "GL_LINE_STIPPLE";
    case GL_POLYGON_OFFSET_FILL: return "GL_POLYGON_OFFSET_FILL";
    case GL_POLYGON_OFFSET_LINE: return "GL_POLYGON_OFFSET_LINE";
    case GL_POLYGON_OFFSET_POINT: return "GL_POLYGON_OFFSET_POINT";
    case GL_COLOR_SUM: return "GL_COLOR_SUM";
    case GL_TEXTURE_1D: return "GL_TEXTURE_1D";
    case GL_TEXTURE_2D: return "GL_TEXTURE_2D";
    case GL_TEXTURE_3D: return "GL_TEXTURE_3D";

    case GL_TEXTURE_CUBE_MAP: return "GL_TEXTURE_CUBE_MAP";

    case GL_TEXTURE_GEN_Q: return "GL_TEXTURE_GEN_Q";
    case GL_TEXTURE_GEN_R: return "GL_TEXTURE_GEN_R";
    case GL_TEXTURE_GEN_S: return "GL_TEXTURE_GEN_S";
    case GL_TEXTURE_GEN_T: return "GL_TEXTURE_GEN_T";

    case GL_STENCIL_TEST: return "GL_STENCIL_TEST";

    case GL_CLIP_PLANE0: return "GL_CLIP_PLANE0";
    case GL_CLIP_PLANE1: return "GL_CLIP_PLANE1";
    case GL_CLIP_PLANE2: return "GL_CLIP_PLANE2";
    case GL_CLIP_PLANE3: return "GL_CLIP_PLANE3";
    case GL_CLIP_PLANE4: return "GL_CLIP_PLANE4";
    case GL_CLIP_PLANE5: return "GL_CLIP_PLANE5";

    case GL_LIGHT0: return "GL_LIGHT0";
    case GL_LIGHT1: return "GL_LIGHT1";
    case GL_LIGHT2: return "GL_LIGHT2";
    case GL_LIGHT3: return "GL_LIGHT3";
    case GL_LIGHT4: return "GL_LIGHT4";
    case GL_LIGHT5: return "GL_LIGHT5";
    case GL_LIGHT6: return "GL_LIGHT6";
    case GL_LIGHT7: return "GL_LIGHT7";
    default: {
	char s[1024];
	sprintf(s, "0x%x", mode);
	return std::string(s);
    }
    }
}


void EdStateSet::setObject(osg::Object *object) {
    model=dynamic_cast<osg::StateSet *> (object);
    if (model == 0) {
	std::cerr << "EdStateSet::setObject(): FATAL ERROR: This is not an 'osg::StateSet'" << std::endl;
	throw -1;
    }
    modes = new EdList( _("Modes"), _("Mode of each state attribute of the set") );

    osg::StateSet::ModeList mlist = model->getModeList();
    osg::StateSet::ModeList::iterator miter;
    for (miter=mlist.begin();miter != mlist.end();miter++) {
	EditableAttributeList list;
	osg::ref_ptr<EdBool> on;
	osg::ref_ptr<EdBool> override;
	osg::ref_ptr<EdBool> _protected;
	on = new EdBool( _("On"), _("Enabled/Disabled"), miter->second & osg::StateAttribute::ON );

	list.push_back(on.get());
        override = new EdBool( _("Override"), _("This attribute will try to override attributes on child nodes"), (miter->second & osg::StateAttribute::OVERRIDE) != 0 );
	list.push_back(override.get());
	_protected = new EdBool( _("Protected"), _("This attribute can't be overriden by any parent's attribute"), (miter->second & osg::StateAttribute::PROTECTED) != 0 );

	list.push_back(_protected.get());
	modes->add(getGLName(miter->first), list);
    }

    list.push_back(modes.get());
    attributes = new EdList( _("Attributes"), _("List of attributes in the set") );

    EditableAttributeList attr_prototypes;
    EdReference *attribute = new EdReference( _("Attribute"), _("State Attribute") );
    attribute->addClass(_("Material"), new EdMaterial);
    attribute->addClass(_("BlendFunc"), new EdBlendFunc);
    attribute->addClass(_("CullFace"), new EdCullFace);
    attribute->addClass(_("FrontFace"), new EdFrontFace);
    attribute->addClass(_("LightModel"), new EdLightModel);
    attribute->addClass(_("ShadeModel"), new EdShadeModel);
    attribute->addClass(_("Fog"), new EdFog);
    attribute->addClass(_("VertexProgram"), new EdVertexProgram);
    attr_prototypes.push_back(attribute);
    attributes->allowAddRemove(attr_prototypes);

    osg::StateSet::AttributeList alist = model->getAttributeList();
    osg::StateSet::AttributeList::iterator aiter;
    for (aiter=alist.begin();aiter != alist.end();aiter++) {
	EditableAttributeList list;
	osg::ref_ptr<EdReference> my_attribute = (EdReference*)attribute->clone();
	my_attribute->setInitial(EntityFactory::instance()->findEntity(aiter->second.first.get()));

	list.push_back(my_attribute.get());
	attributes->add(aiter->second.first->className(),list);
    }

    list.push_back(attributes.get());
    texture_modes = new EdList( _("TextureModes"), _("Mode of each texture state attribute of the set") );

    osg::StateSet::TextureModeList tmlist = model->getTextureModeList();
    osg::StateSet::ModeList::iterator tmiter;
    for (unsigned unit = 0; unit < tmlist.size(); unit++) {
	for (tmiter=tmlist[unit].begin();tmiter != tmlist[unit].end();tmiter++) {
	    EditableAttributeList list;
	    osg::ref_ptr<EdBool> on;
	    osg::ref_ptr<EdBool> override;
	    osg::ref_ptr<EdBool> _protected;
	    on = new EdBool( _("On"), _("Enabled/Disabled"), tmiter->second & osg::StateAttribute::ON );

	    list.push_back(on.get());
	    override = new EdBool( _("Override"), _("This texture attribute will try to override attributes on child nodes"), (tmiter->second & osg::StateAttribute::OVERRIDE) != 0 );

	    list.push_back(override.get());
	    _protected = new EdBool( _("Protected"), _("This texture attribute can't be overriden by any parent's attribute"), (tmiter->second & osg::StateAttribute::PROTECTED) != 0 );

	    list.push_back(_protected.get());
	    char s[1024];
            sprintf(s,"%s<%d>",getGLName(tmiter->first).c_str(),unit);
	    texture_modes->add(s,list);
	}
    }


    list.push_back(texture_modes.get());
    tex_attributes = new EdList( _("Texture Attributes"), _("List of texture attributes in the set") );

    EditableAttributeList tex_attr_prototypes;
    EdInt *tex_unit = new EdInt( _("OpenGL texture unit"), _("The texture unit of OpenGL where this attribute will have effect"), 0, 0, 65536 );
    tex_attr_prototypes.push_back(tex_unit);
    EdReference *tex_attribute = new EdReference( _("Attribute"), _("Texture State Attribute") );
    tex_attribute->addClass(_("Texture2D"), new EdTexture);
    tex_attribute->addClass(_("TexMat"), new EdTexMat);
    tex_attr_prototypes.push_back(tex_attribute);
    tex_attributes->allowAddRemove(tex_attr_prototypes);


    osg::StateSet::TextureAttributeList talist = model->getTextureAttributeList();
    osg::StateSet::AttributeList::iterator taiter;
    for (unsigned unit = 0; unit < tmlist.size(); unit++) {
	for (taiter=talist[unit].begin();taiter != talist[unit].end();taiter++) {
	    EditableAttributeList list;
	    osg::ref_ptr<EdInt> texture_unit;
	    osg::ref_ptr<EdReference> attribute;
	    texture_unit = (EdInt*)tex_unit->clone();
            texture_unit->setInitial(unit);
	    list.push_back(texture_unit.get());

            attribute=(EdReference*)tex_attribute->clone();
	    attribute->setInitial(EntityFactory::instance()->findEntity(taiter->second.first.get()));
	    list.push_back(attribute.get());
	    char s[1024];
            sprintf(s,"%s<%d>",taiter->second.first->className(),unit);
	    tex_attributes->add(s,list);
	}
    }
    list.push_back(tex_attributes.get());
    rendering_hint = new EdEnum( _("Rendering hint"), _("Specify here in which renderbin will this stateset fall") );
    rendering_hint->addOption(_("Default"),osg::StateSet::DEFAULT_BIN);
    rendering_hint->addOption(_("Opaque"),osg::StateSet::OPAQUE_BIN);
    rendering_hint->addOption(_("Transparent"),osg::StateSet::TRANSPARENT_BIN);
    rendering_hint->setInitial(model->getRenderingHint());

    list.push_back(rendering_hint.get());

}

EdStateSet::EdStateSet() {}

EdStateSet::EdStateSet(osg::StateSet *_model) {
    setObject(_model);
}

Command *EdStateSet::compile() {
    osg::StateSet::ModeList _modes;
    unsigned i;
    osg::StateSet::ModeList mlist = model->getModeList();
    osg::StateSet::ModeList::iterator miter;
    for (miter=mlist.begin(),i=0;miter != mlist.end();miter++,i++) {
	std::cout << "i: " << i << std::endl;
	const EditableAttributeList &list =modes->get(i);
        const EdBool *ed_on = dynamic_cast<const EdBool*>(list[0].get());
	bool is_on = ed_on->get();
        const EdBool *ed_override = dynamic_cast<const EdBool*>(list[1].get());
	bool is_override = ed_override->get();
        const EdBool *ed_protected = dynamic_cast<const EdBool*>(list[2].get());
	bool is_protected = ed_protected->get();
	osg::StateAttribute::GLModeValue value=is_on*osg::StateAttribute::ON
	    | is_override*osg::StateAttribute::OVERRIDE
	    | is_protected*osg::StateAttribute::PROTECTED;

	_modes[miter->first]=value;
    }

    ComStateSet::AttributeList _attributes;
    for (unsigned i=0;i<attributes->length();i++) {
	EdReference *ref = dynamic_cast<EdReference*>(attributes->get(i,0));
	osg::StateAttribute *attr = dynamic_cast<osg::StateAttribute*>(ref->get()->getObject());
	_attributes.push_back(attr);
    }


    osg::StateSet::TextureModeList tmlist = model->getTextureModeList();
    osg::StateSet::ModeList::iterator tmiter;

    osg::StateSet::TextureModeList _texture_modes(tmlist.size());

    for (unsigned unit = 0,i=0; unit < tmlist.size(); unit++) {
	for (tmiter=tmlist[unit].begin();tmiter != tmlist[unit].end();tmiter++,i++) {
	    const EditableAttributeList &list =texture_modes->get(i);
	    const EdBool *ed_on = dynamic_cast<const EdBool*>(list[0].get());
	    bool is_on = ed_on->get();
	    const EdBool *ed_override = dynamic_cast<const EdBool*>(list[1].get());
	    bool is_override = ed_override->get();
	    const EdBool *ed_protected = dynamic_cast<const EdBool*>(list[2].get());
	    bool is_protected = ed_protected->get();
	    osg::StateAttribute::GLModeValue value=is_on*osg::StateAttribute::ON
		| is_override*osg::StateAttribute::OVERRIDE
		| is_protected*osg::StateAttribute::PROTECTED;

	    _texture_modes[unit][tmiter->first]=value;
	}
    }

    ComStateSet::TextureAttributeList _tex_attributes;
    for (unsigned i=0;i<tex_attributes->length();i++) {
	EdInt *edunit = dynamic_cast<EdInt*>(tex_attributes->get(i,0));
        size_t unit = edunit->get();
	EdReference *ref = dynamic_cast<EdReference*>(tex_attributes->get(i,1));
	osg::StateAttribute *attr = dynamic_cast<osg::StateAttribute*>(ref->get()->getObject());
	while (unit >= _tex_attributes.size()) {
	    _tex_attributes.push_back(ComStateSet::AttributeList());
	}
	_tex_attributes[unit].push_back(attr);
    }

    return new ComStateSet(model.get(),
			_modes,
                        _attributes,
			_texture_modes,
                        _tex_attributes,
			(int)rendering_hint->get());
}


void EdStateSet::commit() {
    osg::ref_ptr<Command> com = compile();
    com->execute();
    reloadModes();
}

void EdStateSet::reloadModes() {
    modes->clear();
    texture_modes->clear();

    osg::StateSet::ModeList mlist = model->getModeList();
    osg::StateSet::ModeList::iterator miter;
    for (miter=mlist.begin();miter != mlist.end();miter++) {
	EditableAttributeList list;
	osg::ref_ptr<EdBool> on;
	osg::ref_ptr<EdBool> override;
	osg::ref_ptr<EdBool> _protected;
	on = new EdBool( _("On"), _("Enabled/Disabled"), miter->second & osg::StateAttribute::ON );

	list.push_back(on.get());
	override = new EdBool( _("Override"), _("This attribute will try to override attributes on child nodes"), (miter->second & osg::StateAttribute::OVERRIDE) != 0 );

	list.push_back(override.get());
	_protected = new EdBool( _("Protected"), _("This attribute can't be overriden by any parent's attribute"), (miter->second & osg::StateAttribute::PROTECTED) != 0 );

	list.push_back(_protected.get());
	modes->add(getGLName(miter->first), list);
    }

    osg::StateSet::TextureModeList tmlist = model->getTextureModeList();
    osg::StateSet::ModeList::iterator tmiter;
    for (unsigned unit = 0; unit < tmlist.size(); unit++) {
	for (tmiter=tmlist[unit].begin();tmiter != tmlist[unit].end();tmiter++) {
	    EditableAttributeList list;
	    osg::ref_ptr<EdBool> on;
	    osg::ref_ptr<EdBool> override;
	    osg::ref_ptr<EdBool> _protected;
	    on = new EdBool( _("On"), _("Enabled/Disabled"), tmiter->second & osg::StateAttribute::ON );

	    list.push_back(on.get());
	    override = new EdBool( _("Override"), _("This texture attribute will try to override attributes on child nodes"), (tmiter->second & osg::StateAttribute::OVERRIDE) != 0 );

	    list.push_back(override.get());
	    _protected = new EdBool( _("Protected"), _("This texture attribute can't be overriden by any parent's attribute"), (tmiter->second & osg::StateAttribute::PROTECTED) != 0 );

	    list.push_back(_protected.get());
	    char s[1024];
            sprintf(s,"%s<%d>",getGLName(tmiter->first).c_str(),unit);
	    texture_modes->add(s,list);
	}
    }
}
